﻿using Microsoft.EntityFrameworkCore;
using StatisticsService.Api.Extensions.MappingExtensions;
using StatisticsService.Database.Models;
using StatisticsService.Domain.Filters;
using StatisticsService.Dto;

namespace StatisticsService.Api.Extensions.QueryExtensions
{
    public static class QueryExtensions
    {
        public static async Task<IEnumerable<ClientDto>> GetDailyActiveClientsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilter filter)
        {
            return await query
                .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
                .Select(x => new { x.TopicId, x.EventTime })
                .Distinct()
                .GroupBy(x => x.EventTime.Date)
                .Select(x => new ClientDto { Date = x.Key, Count = x.Count() })
                .OrderBy(x => x.Date)
                .ToListAsync();
        }

        public static async Task<IEnumerable<ClientDto>> GetDailyActiveClientsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilterRID filter)
        {
            return await query
                .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
                .GroupBy(x => x.EventTime)
                .Select(x => new ClientDto { Date = x.Key, Count = x.Count() })
                .OrderBy(x => x.Date)
                .ToListAsync();
        }

        public static async Task<int> GetTotalActiveClientsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilter filter)
        {
            return await query
            .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
            .Select(x => new { x.TopicId })
            .Distinct()
            .CountAsync();
        }

        public static async Task<int> GetTotalActiveClientsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilterRID filter)
        {
            return await query
            .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
            .CountAsync();
        }

        public static async Task<IEnumerable<AccountDto>> GetDailyActiveAccountsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilter filter)
        {
            return await query
                .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
                .Select(x => new { x.TopicId, x.EventTime.Date }) //TODO: Replace TopicId with AccountNumber when account numbers are added to model
                .Distinct()
                .GroupBy(x => x.Date)
                .Select(x => new AccountDto { Date = x.Key, Count = x.Count() })
                .OrderBy(x => x.Date)
                .ToListAsync();
        }

        public static async Task<int> GetTotalActiveAccountsFromQueryAsync(this IQueryable<AggregatedEventDay> query, BaseFilter filter)
        {
            return await query
            .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
            .Select(x => new { x.TopicId }) //TODO: Replace TopicId with AccountNumber when account numbers are added to model
            .Distinct()
            .CountAsync();
        }

        // TODO: create async version when source is not an enumerable.
        // Casting with .AsQueryable does not implement proper interfaces either.
        public static List<ActiveUserDto> GetDailyActiveUsersFromQuery(this IQueryable<User> query, BaseFilter filter)
        {
            return query
            .Where(x => query.ApplyFilterUser(filter).Contains(x.Timestamp.Date))
            .Select(x => new { x.Id, x.Timestamp.Date })
            .Distinct()
            .GroupBy(x => x.Date)
            .Select(x => new ActiveUserDto() { Date = x.Key, Count = x.Count() })
            .OrderBy(x => x.Date)
            .ToList();
        }

        public static int GetTotalActiveUsersFromQuery(this IQueryable<User> query, BaseFilter filter)
        {
            return query
            .Where(x => query.ApplyFilterUser(filter).Contains(x.Timestamp.Date))
            .Select(x => new { x.Id })
            .Distinct()
            .Count();
        }

        // TODO create async version when source is not an enumerable.
        // Casting with .AsQueryable does not implement proper interfaces either.
        public static List<UserDto> GetUsersFromQuery(this IQueryable<User> query, BaseFilter filter)
        {
            return query
                .Where(x => query.ApplyFilterUser(filter).Contains(x.Timestamp.Date))
                .Select(x => x.ToUserDto())
                .OrderBy(x => x.Timestamp)
                .ToList();
        }

        public static async Task<List<StatisticsDto>> GetStatisticsFromQueryAsync(this IQueryable<AggregatedEventBase> query, StatisticsFilter filter)
        {
            filter.ToDate = filter.TimeType == "hour" ? filter.ToDate.AddDays(1) : filter.ToDate;

            return await query
                .FilterByDateRange(filter)
                .OrderByDescending(x => x.EventTime)
                .SkipIfNotNull(filter.Skip)
                .TakeIfNotNull(filter.Take)
                .OrderBy(x => x.EventTime)
                .Select(x => x.ToStatisticsDto())
                .ToListAsync();
        }
        public static async Task<List<T>> GetAggregatedStatisticsFromQueryAsync<T>(this IQueryable<AggregatedEventBase> query, StatisticsFilter filter) where T : BaseActiveResourceDto, new()
        {
            filter.ToDate = filter.TimeType == "hour" ? filter.ToDate.AddDays(1) : filter.ToDate;

            return await query
                .Where(x => query.ApplyFilter(filter).Contains(x.EventTime.Date))
                .GroupBy(x => x.EventTime)
                .Select(x => new T() { Date = x.Key, Count = x.Count() + (int)x.Sum(y => y.Count) + (int)x.Sum(y => y.EventCount - 1) })
                .OrderBy(x => x.Date)
                .ToListAsync();
        }

        /// <summary>
        /// Common method for all repository methods using date ranges (Methods including BaseFilter or derivatives)
        /// Used to ensure date filtering behaviour.
        /// </summary>
        /// <param name="query"></param>
        /// <param name="filter"></param>
        /// <returns></returns>
        public static IQueryable<AggregatedEventBase> FilterByDateRange(this IQueryable<AggregatedEventBase> query, BaseFilter filter)
        {
            return query.Where(x => x.EventTime >= filter.FromDate && x.EventTime <= filter.ToDate);
        }

        /// <summary>
        /// Applies all Basefilter values to query (Date range, skip and take)
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="query"></param>
        /// <param name="filter"></param>
        public static IQueryable<DateTime> ApplyFilter<T>(this IQueryable<T> query, BaseFilter filter) where T : AggregatedEventBase
        {
            return query
                .FilterByDateRange(filter)
                .Select(x => x.EventTime)
                .Distinct()
                .OrderByDescending(x => x)
                .SkipIfNotNull(filter.Skip)
                .TakeIfNotNull(filter.Take)
                .OrderBy(x => x);
        }

        /// <summary>
        /// Does the same thing as ApplyFilter, but for user objects.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="query"></param>
        /// <param name="filter"></param>
        public static IQueryable<DateTime> ApplyFilterUser<T>(this IQueryable<T> query, BaseFilter filter) where T : User
        {
            return query
                .Where(x => x.Timestamp >= filter.FromDate && x.Timestamp <= filter.ToDate)
                .Select(x => x.Timestamp.Date)
                .OrderByDescending(x => x)
                .SkipIfNotNull(filter.Skip)
                .TakeIfNotNull(filter.Take)
                .OrderBy(x => x);
        }

        public static IQueryable<T> TakeIfNotNull<T>(this IQueryable<T> source, int? count)
        {
            return !count.HasValue ? source : source.Take(count.Value);
        }

        public static IQueryable<T> SkipIfNotNull<T>(this IQueryable<T> source, int? count)
        {
            return !count.HasValue ? source : source.Skip(count.Value);
        }
    }
}
