﻿using Microsoft.EntityFrameworkCore;
using StatisticsService.Api.Extensions.QueryExtensions;
using StatisticsService.Database;
using StatisticsService.Database.Models;
using StatisticsService.Domain.Filters;
using StatisticsService.Dto;
using StatisticsService.Interfaces.Repository;
using ZData.AccessServer.Client;

namespace StatisticsService.Api.Repository
{
    public class UsersRepository : BaseRepository, IUsersRepository
    {
        /// <summary>
        /// Constructor
        /// </summary>
        public UsersRepository(StatisticsServiceDbContext context) : base(context) { }
        public UsersRepository(StatisticsServiceDbContext context, IZamManagementClient zamManagementClient) : base(context, zamManagementClient) { }

        /// <summary>
        /// Returns all users
        /// </summary>
        /// <param name="filter">A filter containing relevant data for query</param>
        /// <returns>List of all users</returns>
        public async Task<IEnumerable<UserDto>> GetAllUsers(BaseFilter filter)
        {
            return Users.AsQueryable()
                .GetUsersFromQuery(filter);
        }

        /// <summary>
        /// Returns all users by resource
        /// </summary>
        /// <param name="filter">A filter containing relevant data for query</param>
        /// <returns>List of all users associated with the requested resource id</returns>
        public async Task<IEnumerable<UserDto>> GetUsersByResourceId(BaseFilterRID filter)
        {
            var resourceIds = await GetResourceChildrensByResourceId(filter.ResourceId, filter.ResourceType);

            return Users.AsQueryable().Where(x => resourceIds.Any(resourceId => resourceId == x.ResourceId))
                .GetUsersFromQuery(filter);
        }

        /// <summary>
        /// Creates a list of active users, sorted by date.
        /// </summary>
        /// <param name="filter">A filter containing relevant data for query</param>
        /// <returns>Returns a list of new users per day and a meta object containing total count for period.</returns>
        public async Task<ActiveResourceDto> GetActiveUsersByResourceId(BaseFilterRID filter)
        {
            var resourceIds = await GetResourceChildrensByResourceId(filter.ResourceId, filter.ResourceType);

            var daily = Users
                .Where(x => resourceIds.Any(resourceId => resourceId == x.ResourceId))
                .AsQueryable()
                .GetDailyActiveUsersFromQuery(filter);

            var total = Users
                .Where(x => resourceIds.Any(resourceId => resourceId == x.ResourceId))
                .AsQueryable()
                .GetTotalActiveUsersFromQuery(filter);

            return new ActiveResourceDto(daily, total);
        }

        /// <summary>
        /// Creates a list of active clients, sorted by date.
        /// </summary>
        /// <param name="filter">A filter containing relevant data for query</param>
        /// <returns>Returns a list of new users per day and a meta object containing total count for period. </returns>
        public async Task<ActiveResourceDto> GetAllActiveUsers(BaseFilter filter)
        {
            var daily = Users
                .AsQueryable()
                .GetDailyActiveUsersFromQuery(filter);

            var total = Users
                .AsQueryable()
                .GetTotalActiveUsersFromQuery(filter);

            return new ActiveResourceDto(daily, total);
        }
    }
}
