﻿using IdentityServer4.AccessTokenValidation;
using Microsoft.AspNetCore.Authorization;
using Microsoft.Net.Http.Headers;
using StatisticsService.Api.Swagger;
using ZData.Authentication.Extensions;
using MediatR;
using StatisticsService.Domain;
using ZData.Iam.Enforcer.Client;
using StatisticsService.Common;
using StatisticsService.Api.Repository;
using StatisticsService.Interfaces.Repository;
using StatisticsService.Database;
using Microsoft.EntityFrameworkCore;
using StatisticsService.Api.Extensions;
using StatisticsService.Api.Clients;

namespace StatisticsService.Api
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            services.AddDbContext<StatisticsServiceDbContext>(options =>
            {
                options.UseNpgsql(Configuration.GetConnectionString("DefaultConnection"), 
                    providerOptions => providerOptions.EnableRetryOnFailure());
            }, ServiceLifetime.Transient);

            services.AddMemoryCache();

            services.AddControllers();

            services.AddSwagger(Configuration);
            services.AddMediatR(typeof(DomainRoot));

            services.AddTransient<IAccountsRepository, AccountsRepository>();
            services.AddTransient<IClientsRepository, ClientsRepository>();
            services.AddTransient<IResourcesRepository, ResourcesRepository>();
            services.AddTransient<IStatisticsRepository, StatisticsRepository>();
            services.AddTransient<IUsersRepository, UsersRepository>();

            services.AddTransient<ICachedResourceProviderClient, CachedResourceProviderClient>();

            services.AddAccessServerServices(Configuration);
            services.AddAccessControl(Configuration);
            services.AddHttpClients(Configuration);

            // bank service cron job for retrieving users is not ready
            //services.AddWorkerServices(Configuration);

            services.AddGrpcClients(Configuration);

            services.AddApplicationInsightsTelemetry(Configuration["ApplicationInsights:ConnectionString"]);
            services.AddHealthChecks();
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            if (env.IsDevelopment())
            {
                app.UseDeveloperExceptionPage();
            }

            app.UseSwagger();
            app.UseSwaggerUI(c =>
            {
                c.SwaggerEndpoint("/swagger/v1/swagger.json", "Statistics Service v1");
                c.RoutePrefix = "docs";
            });

            app.UseRouting();
            app.UseSentryTracing();
            app.UseCors();

            app.UseAuthentication();
            app.UseAuthorization();

            app.UseHealthChecks("/health");

            app.UseEndpoints(endpoints =>
            {
                endpoints.MapControllers();
            });
        }
    }
}
