﻿using FluentAssertions;
using Microsoft.EntityFrameworkCore;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using StatisticsService.Api.Repository;
using StatisticsService.Database;
using StatisticsService.Database.Models;
using StatisticsService.Domain.Filters;
using StatisticsService.Dto;
using System;
using System.Collections.Generic;
using System.Threading.Tasks;

namespace StatisticsService.Tests
{
    [TestClass]
    public class ClientsRepositoryTests
    {
        private static readonly DateTime now = DateTime.Now.Date;
        private static readonly string duplicateGUID = Guid.NewGuid().ToString();

        [TestMethod]
        public async Task GetAllClients()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-clients")
                .Options;

            var context = new StatisticsServiceDbContext(options);

            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilter();

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-3), 1),
                    new ClientDto(now.AddDays(-2), 1),
                    new ClientDto(now.AddDays(-1), 2),
                    new ClientDto(now, 2)
                };

            var expectedMeta = 4;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetAllActiveClients(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAllClientsSkip()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-clients-skip")
                .Options;

            var context = new StatisticsServiceDbContext(options);

            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilter() { Skip = 2};

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-3), 1),
                    new ClientDto(now.AddDays(-2), 1),
                };

            var expectedMeta = 2;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetAllActiveClients(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAllClientsTake()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-clients")
                .Options;

            var context = new StatisticsServiceDbContext(options);

            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilter() { Take = 2};

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-1), 2),
                    new ClientDto(now, 2)
                };

            var expectedMeta = 3;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetAllActiveClients(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetClientsByResourceId()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-clients-by-resource-id")
                .Options;

            var context = new StatisticsServiceDbContext(options);
            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilterRID() { ResourceId = duplicateGUID };

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-3), 2),
                    new ClientDto(now.AddDays(-1), 2),
                    new ClientDto(now, 1)
                };

            var expectedMeta = 5;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetActiveClientsByResourceId(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetClientsByResourceIdSkip()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-clients-by-resource-id-skip")
                .Options;

            var context = new StatisticsServiceDbContext(options);
            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilterRID() { ResourceId = duplicateGUID, Skip = 2 };

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-3), 2),
                };

            var expectedMeta = 2;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetActiveClientsByResourceId(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetClientsByResourceIdTake()
        {
            //Arrange 
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-clients-by-resource-id-take")
                .Options;

            var context = new StatisticsServiceDbContext(options);
            await context.AggregatedEventsDay.AddRangeAsync(TestEvents());
            await context.SaveChangesAsync();

            var clientsRepository = new ClientsRepository(context);
            var filter = new BaseFilterRID() { ResourceId = duplicateGUID, Take = 2 };

            var expectedData = new List<ClientDto>()
                {
                    new ClientDto(now.AddDays(-1), 2),
                    new ClientDto(now, 1)
                };

            var expectedMeta = 3;

            ActiveResourceDto expected = new(expectedData, expectedMeta);

            //Act
            var actual = await clientsRepository.GetActiveClientsByResourceId(filter);

            //Assert
            actual.Should().NotBeNull().And
                .BeOfType<ActiveResourceDto>().And
                .BeEquivalentTo(expected);

            //Cleanup
            context.Database.EnsureDeleted();
        }

        private static AggregatedEventDay An_Aggregated_Event_Day(int id, string topicId, DateTime eventTime)
        {
            return new AggregatedEventDay()
            {
                Id = id,
                TopicId = topicId,
                ResourceType = "company",
                EventTime = eventTime.Date,
                EventType = "ZData.Events.Bank.ReceivableReceived",
                EventCount = 100,
                Count = 4,
            };
        }

        private List<AggregatedEventDay> TestEvents()
        {
            return new List<AggregatedEventDay>()
            {
            An_Aggregated_Event_Day(1, Guid.NewGuid().ToString(), now),
            An_Aggregated_Event_Day(2, duplicateGUID, now),
            An_Aggregated_Event_Day(3, Guid.NewGuid().ToString(), now.AddDays(-1)),
            An_Aggregated_Event_Day(4, duplicateGUID, now.AddDays(-1)),
            An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1)),
            An_Aggregated_Event_Day(6, Guid.NewGuid().ToString(), now.AddDays(-2)),
            An_Aggregated_Event_Day(7, duplicateGUID, now.AddDays(-3)),
            An_Aggregated_Event_Day(8, duplicateGUID, now.AddDays(-3)),
            An_Aggregated_Event_Day(9, Guid.NewGuid().ToString(), now.AddDays(-8)),
            An_Aggregated_Event_Day(10, duplicateGUID, now.AddDays(-8))
            };
        }
    }
}