﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using FluentAssertions;
using Microsoft.EntityFrameworkCore;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using StatisticsService.Api.Repository;
using StatisticsService.Database;
using StatisticsService.Database.Models;
using StatisticsService.Domain.Filters;
using StatisticsService.Dto;

namespace StatisticsService.Tests
{
    [TestClass]
    public class StatisticsRepositoryTests
    {
        private static readonly DateTime now = DateTime.Now.Date;
        private static readonly string duplicateGUID = Guid.NewGuid().ToString();
        private const string StatementsReceivedFull = "ZData.Events.Bank.StatementsReceived";
        private const string ReceivablesReceivedFull = "ZData.Events.Bank.ReceivableReceived";
        private const string ReceivablesReceived = "Receivables";
        private const string PaymentsSentToBankFull = "ZData.Events.Bank.PaymentsSentToBank";

        [TestMethod]
        public async Task GetAllStatistics_ByEventType_ReturnsAllStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, Guid.NewGuid().ToString(), now, 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(2, duplicateGUID, now, 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(3, Guid.NewGuid().ToString(), now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(4, duplicateGUID, now.AddDays(-1), 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(6, Guid.NewGuid().ToString(), now.AddDays(-2), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(7, duplicateGUID, now.AddDays(-3), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(8, duplicateGUID, now.AddDays(-3), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(9, Guid.NewGuid().ToString(), now.AddDays(-8), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(10, duplicateGUID, now.AddDays(-8), 1, 1, PaymentsSentToBankFull)
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived,
            };

            // Act
            var actual = await statisticsRepository.GetStatistics(filter);

            // Assert
            actual.Should().NotBeNull().And
                .HaveCount(4).And
                .AllBeOfType<StatisticsDto>();

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAllStatistics_ByEventTypeSkip_ReturnsAllStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics-by-resourceid")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, Guid.NewGuid().ToString(), now, 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(2, duplicateGUID, now, 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(3, Guid.NewGuid().ToString(), now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(4, duplicateGUID, now.AddDays(-1), 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(6, Guid.NewGuid().ToString(), now.AddDays(-2), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(7, duplicateGUID, now.AddDays(-3), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(8, duplicateGUID, now.AddDays(-3), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(9, Guid.NewGuid().ToString(), now.AddDays(-8), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(10, duplicateGUID, now.AddDays(-8), 1, 1, PaymentsSentToBankFull)
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived,
                Skip = 2,
            };

            // Act
            var actual = await statisticsRepository.GetStatistics(filter);

            // Assert
            actual.Should().NotBeNull().And
                .HaveCount(2).And
                .AllBeOfType<StatisticsDto>();

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAllStatistics_ByEventTypeTake_ReturnsAllStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics-by-resourceid")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, Guid.NewGuid().ToString(), now, 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(2, duplicateGUID, now, 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(3, Guid.NewGuid().ToString(), now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(4, duplicateGUID, now.AddDays(-1), 1, 1, PaymentsSentToBankFull),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(6, Guid.NewGuid().ToString(), now.AddDays(-2), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(7, duplicateGUID, now.AddDays(-3), 1, 1, StatementsReceivedFull),
                An_Aggregated_Event_Day(8, duplicateGUID, now.AddDays(-3), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(9, Guid.NewGuid().ToString(), now.AddDays(-8), 1, 1, ReceivablesReceivedFull),
                An_Aggregated_Event_Day(10, duplicateGUID, now.AddDays(-8), 1, 1, PaymentsSentToBankFull)
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived,
                Take = 2,
            };

            // Act
            var actual = await statisticsRepository.GetStatistics(filter);

            // Assert
            actual.Should().NotBeNull().And
                .HaveCount(2).And
                .AllBeOfType<StatisticsDto>();

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetStatisticsByResource_ValidResourceId_ReturnsStatisticsByResource()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics-by-resourceid")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var filter = new StatisticsFilterRID()
            {
                ResourceId = duplicateGUID,
                EventType = ReceivablesReceived,
            };

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var expected = new List<StatisticsDto>()
            {
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now.AddDays(-2), 3, 120, null),
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now.AddDays(-1), 7, 50, null),
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now, 1, 100, null),
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now, 1, 100, null),
            };

            // Act
            var statisticsRepository = new StatisticsRepository(context);
            var actual = await statisticsRepository.GetStatisticsByResource(filter);

            // Assert
            actual.Should().HaveCount(4).And
                .AllBeOfType<StatisticsDto>().And
                .BeEquivalentTo(expected);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetStatisticsByResource_ValidResourceIdSkip_ReturnsStatisticsByResource()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics-by-resourceid")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var filter = new StatisticsFilterRID()
            {
                ResourceId = duplicateGUID,
                EventType = ReceivablesReceived,
                Skip = 2,
            };

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var expected = new List<StatisticsDto>()
            {
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now.AddDays(-2), 3, 120, null),
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now.AddDays(-1), 7, 50, null),
            };

            // Act
            var statisticsRepository = new StatisticsRepository(context);
            var actual = await statisticsRepository.GetStatisticsByResource(filter);

            // Assert
            actual.Should().HaveCount(2).And
                .AllBeOfType<StatisticsDto>().And
                .BeEquivalentTo(expected);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetStatisticsByResource_ValidResourceIdTake_ReturnsStatisticsByResource()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-all-statistics-by-resourceid")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var filter = new StatisticsFilterRID()
            {
                ResourceId = duplicateGUID,
                EventType = ReceivablesReceived,
                Take = 2
            };

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AggregatedEventsDay.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var expected = new List<StatisticsDto>()
            {
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now, 1, 100, null),
                new StatisticsDto(duplicateGUID, ReceivablesReceivedFull, now, 1, 100, null),
            };

            // Act
            var statisticsRepository = new StatisticsRepository(context);
            var actual = await statisticsRepository.GetStatisticsByResource(filter);

            // Assert
            actual.Should().HaveCount(2).And
                .AllBeOfType<StatisticsDto>().And
                .BeEquivalentTo(expected);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAggregatedStatistics_SameDay_ReturnsAggregatedStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-aggregated-statistics")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = Aggregated_Event_Days(4, true);

            await context.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived
            };

            var aggregatedStatistics = new List<AggregatedStatisticsDto>
            {
                new AggregatedStatisticsDto() { Count = 404, Date = now }
            };

            // Act
            var actual = await statisticsRepository.GetAggregatedStatistics(filter);

            // Assert
            actual.Should().NotBeNull().And
                .AllBeOfType<AggregatedStatisticsDto>().And
                .BeEquivalentTo(aggregatedStatistics);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAggregatedStatistics_MultipleEvents_MultipleDays_ReturnsAggregatedStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-aggregated-statistics")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived
            };

            var aggregatedStatistics = new List<AggregatedStatisticsDto>()
            {
                new AggregatedStatisticsDto { Count = 107, Date = now.AddDays(-3) },
                new AggregatedStatisticsDto { Count = 123, Date = now.AddDays(-2) },
                new AggregatedStatisticsDto { Count = 57, Date = now.AddDays(-1) },
                new AggregatedStatisticsDto { Count = 303, Date = now },
            };

            // Act
            var actual = await statisticsRepository.GetAggregatedStatistics(filter);

            // Assert
            actual.Should().HaveCount(4).And
                .AllBeOfType<AggregatedStatisticsDto>().And
                .BeEquivalentTo(aggregatedStatistics);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAggregatedStatisticsByResourceId_ValidResourceId_ReturnsAggregatedStatisticsByResourceId()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
            .UseInMemoryDatabase(databaseName: "test-get-aggregated-statistics")
            .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
                An_Aggregated_Event_Day(7, Guid.NewGuid().ToString(), now.AddDays(-2)),
            };

            await context.AddRangeAsync(events);
            await context.SaveChangesAsync();
            
            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilterRID()
            {
                ResourceId = duplicateGUID,
                EventType = ReceivablesReceived
            };

            var aggregatedStatistics = new List<AggregatedStatisticsDto>()
            {
                new AggregatedStatisticsDto { Count = 123, Date = now.AddDays(-2) },
                new AggregatedStatisticsDto { Count = 57, Date = now.AddDays(-1) },
                new AggregatedStatisticsDto { Count = 202, Date = now },
            };

            // Act
            var actual = await statisticsRepository.GetAggregatedStatisticsByResource(filter);

            // Assert
            actual.Should().HaveCount(3).And
                .AllBeOfType<AggregatedStatisticsDto>().And
                .BeEquivalentTo(aggregatedStatistics);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAggregatedStatistics_Skip_ReturnsAggregatedStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-aggregated-statistics")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived,
                Skip = 2
            };

            var aggregatedStatistics = new List<AggregatedStatisticsDto>()
            {
                new AggregatedStatisticsDto { Count = 107, Date = now.AddDays(-3) },
                new AggregatedStatisticsDto { Count = 123, Date = now.AddDays(-2) },
            };

            // Act
            var actual = await statisticsRepository.GetAggregatedStatistics(filter);

            // Assert
            actual.Should().HaveCount(2).And
                .AllBeOfType<AggregatedStatisticsDto>().And
                .BeEquivalentTo(aggregatedStatistics);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        [TestMethod]
        public async Task GetAggregatedStatistics_Take_ReturnsAggregatedStatistics()
        {
            // Arrange
            var options = new DbContextOptionsBuilder<StatisticsServiceDbContext>()
                .UseInMemoryDatabase(databaseName: "test-get-aggregated-statistics")
                .Options;
            var context = new StatisticsServiceDbContext(options);

            var events = new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(1, duplicateGUID, now),
                An_Aggregated_Event_Day(2, Guid.NewGuid().ToString(), now.AddDays(-3), 102, 5),
                An_Aggregated_Event_Day(3, duplicateGUID, now),
                An_Aggregated_Event_Day(4, Guid.NewGuid().ToString(), now),
                An_Aggregated_Event_Day(5, duplicateGUID, now.AddDays(-1), 50, 7),
                An_Aggregated_Event_Day(6, duplicateGUID, now.AddDays(-2), 120, 3),
            };

            await context.AddRangeAsync(events);
            await context.SaveChangesAsync();

            var statisticsRepository = new StatisticsRepository(context);
            var filter = new StatisticsFilter()
            {
                EventType = ReceivablesReceived,
                Take = 2
            };

            var aggregatedStatistics = new List<AggregatedStatisticsDto>()
            {
                new AggregatedStatisticsDto { Count = 57, Date = now.AddDays(-1) },
                new AggregatedStatisticsDto { Count = 303, Date = now },
            };

            // Act
            var actual = await statisticsRepository.GetAggregatedStatistics(filter);

            // Assert
            actual.Should().HaveCount(2).And
                .AllBeOfType<AggregatedStatisticsDto>().And
                .BeEquivalentTo(aggregatedStatistics);

            // Cleanup
            context.Database.EnsureDeleted();
        }

        private static AggregatedEventDay An_Aggregated_Event_Day(int id, string topicId, DateTime eventTime, int with_eventcount = 100, int with_count = 1, string type = ReceivablesReceivedFull)
        {
            return new AggregatedEventDay()
            {
                Id = id,
                TopicId = topicId,
                ResourceType = "company",
                EventTime = eventTime,
                EventType = type,
                EventCount = with_eventcount,
                Count = with_count,
            };
        }

        private static List<AggregatedEventDay> Aggregated_Event_Days(int count, bool sameDay = false)
        {
            var list = new List<AggregatedEventDay>();
            for (var i = 0; i < count; i++)
            {
                Guid guid = Guid.NewGuid();
                list.Add(An_Aggregated_Event_Day(i + 1, guid.ToString(), sameDay ? now : now.AddDays(-i)));
            }
            return list;
        }

        /*
        private List<AggregatedEventDay> List_of_Aggregated_Event_Day()
        {
            return new List<AggregatedEventDay>()
            {
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
                An_Aggregated_Event_Day(),
            };
        }
        */
    }
}
