import { useEffect, useState } from 'react'
import { useNavigate, useLocation } from 'react-router-dom'
import { SingleValue } from 'react-select'
import Select from 'react-select-virtualized'

import { capitalize } from 'utils'
import { useResources } from 'utils/hooks/useResources'
import ResourceType from 'shared/enums/resource-type.enum'
import ResourceOption from 'shared/interfaces/resource-option.interface'
import Resource from 'shared/interfaces/resources/resource.interface'
import GroupedOption from 'shared/interfaces/grouped-option.interface'

const CustomGroupHeading = (data: GroupedOption) => {
  return (
    <div className="d-flex justify-content-between resource-select-grouped">
      <span>{data.label}</span>
      <span className="badge bg-primary fs-12">{data.options.length}</span>
    </div>
  )
}

const formatResourceAsOption = (resource: Resource): ResourceOption => {
  return {
    label: resource.name.trim(),
    type: resource.type,
    value: resource.id,
  }
}

const formatResourceOptions = (resources: Resource[] | undefined): ResourceOption[] | undefined =>
  resources?.map(resource => formatResourceAsOption(resource))

const getResourceByPath = (
  path: string,
  companies: Resource[] | undefined,
  partners: Resource[] | undefined,
): ResourceOption | null => {
  const resourceId = path.split('/').pop()
  const resource =
    partners?.find(c => c.id === resourceId) || companies?.find(c => c.id === resourceId)

  return resource ? formatResourceAsOption(resource) : null
}

export const ResourceSelectInput = () => {
  const navigate = useNavigate()
  const location = useLocation()
  const [isLoading, setIsLoading] = useState(true)

  const [selectedValue, setSelectedValue] = useState<ResourceOption | null>(null)
  const [groupedOptions, setGroupedOptions] = useState<GroupedOption[]>([])

  const companyQuery = useResources('company')
  const partnerQuery = useResources('partner')

  useEffect(() => {
    setSelectedValue(getResourceByPath(location.pathname, companyQuery.data, partnerQuery.data))

    setGroupedOptions([
      {
        label: capitalize(ResourceType.partner),
        options: formatResourceOptions(partnerQuery.data) ?? [],
      },
      {
        label: capitalize(ResourceType.company),
        options: formatResourceOptions(companyQuery.data) ?? [],
      },
    ])
  }, [companyQuery.data, partnerQuery.data])

  useEffect(() => {
    setIsLoading(companyQuery.isLoading || partnerQuery.isLoading)
  }, [companyQuery.isLoading, partnerQuery.isLoading])

  const changeHandler = (option: SingleValue<ResourceOption>) => {
    setSelectedValue(option)

    const currentProject = location.pathname.split('/').slice(0, 3).join('/')
    if (option) navigate(`${currentProject}/${option?.type}/${option?.value}`)
    else navigate(`${currentProject}`)
  }

  useEffect(() => {
    const lastItem = location.pathname.split('/').pop()
    lastItem != selectedValue?.value && setSelectedValue(null)
  }, [location])

  return (
    <Select
      className="resource-select-container"
      classNamePrefix="resource-select-"
      options={groupedOptions}
      isLoading={isLoading}
      grouped
      placeholder="Type here to search for a resource"
      formatGroupHeaderLabel={CustomGroupHeading}
      groupHeaderHeight={35}
      aria-label="Resource selection menu"
      value={selectedValue}
      onChange={changeHandler}
      noOptionsMessage={() => 'No resources found'}
    />
  )
}
