import { useEffect, useState } from 'react'
import {
  Chart as ChartJS,
  LinearScale,
  LineController,
  LineElement,
  Filler,
  ChartType,
  ScriptableContext,
  ChartOptions,
  ChartData,
  Color,
} from 'chart.js'
import { Chart } from 'react-chartjs-2'
import dayjs, { ManipulateType } from 'dayjs'
import { Settings } from 'react-feather'

import { DatePicker } from '../DatePicker'
import { CardError } from './CardError'
import { GraphSettingsModal } from './GraphSettingsModal'
import { formatNumber } from 'utils'
import { useStatistics } from 'utils/hooks/useStatistics'
import { getFilterTimeType, groupStatisticsByDate } from 'utils/statistics'
import StatisticsParameters from 'shared/interfaces/graph/statistics-parameters.interface'
import TypeChartDataset from 'shared/interfaces/chartjs/charts.interface'
import GroupedDatasets from 'shared/interfaces/chartjs/datasets.interface'
import ResourceOption from 'shared/interfaces/resource-option.interface'

ChartJS.register(LinearScale, LineController, LineElement, Filler)

const CHART_OPTIONS: ChartOptions = {
  interaction: {
    mode: 'index',
    intersect: false,
  },
  maintainAspectRatio: false,
  responsive: true,
  plugins: {
    legend: {
      display: false,
    },
    title: {
      display: false,
    },
    tooltip: {
      displayColors: false,
      intersect: false,
      callbacks: {
        labelTextColor: function (context) {
          return (context?.dataset?.borderColor as Color) ?? ''
        },
        label: function (context) {
          return context.dataset.label
            ? `${context.dataset.label[context.dataIndex]}: ${formatNumber(context.parsed.y)}`
            : ''
        },
      },
    },
  },
  elements: {
    line: {
      tension: 0.3,
    },
  },
  scales: {
    x: {
      display: true,
      grid: {
        display: false,
        drawBorder: false,
        color: 'rgba(119, 119, 142, 0.08)',
      },
      ticks: {
        color: '#b0bac9',
        autoSkip: true,
      },
      title: {
        display: false,
      },
    },
    y: {
      display: true,
      grid: {
        display: true,
        drawBorder: false,
        color: 'rgba(142, 156, 173,0.1)',
      },
      ticks: {
        color: '#b0bac9',
      },
      title: {
        display: false,
      },
    },
  },
}

const getGradient = (context: ScriptableContext<'line'>, variant: string): CanvasGradient => {
  const chart = context.chart
  const { ctx, chartArea } = chart

  let width = 0,
    height = 0,
    gradient = ctx.createLinearGradient(0, 80, 0, 280)

  if (!chartArea) return gradient

  const chartWidth = chartArea.right - chartArea.left
  const chartHeight = chartArea.bottom - chartArea.top
  if (!gradient || width !== chartWidth || height !== chartHeight) {
    width = chartWidth
    height = chartHeight
    gradient = ctx.createLinearGradient(0, chartArea.bottom, 0, chartArea.top)
    gradient.addColorStop(
      0,
      variant === 'primary' ? 'rgba(108, 95, 252, 0.2)' : 'rgba(252,82,150, 0.2)',
    )
    gradient.addColorStop(
      1,
      variant === 'primary' ? 'rgba(108, 95, 252, 0.8)' : 'rgba(252,82,150, 0.8)',
    )
  }

  return gradient
}

const DATASETS_SETTINGS: TypeChartDataset = {
  primary: {
    backgroundColor: (context: ScriptableContext<'line'>) => getGradient(context, 'primary'),
    borderColor: '#6c5ffc',
    pointBackgroundColor: '#6c5ffc',
    pointHoverBackgroundColor: (context: ScriptableContext<'line'>) =>
      getGradient(context, 'primary'),
    pointBorderColor: '#6c5ffc',
    pointHoverBorderColor: (context: ScriptableContext<'line'>) => getGradient(context, 'primary'),
    pointBorderWidth: 2,
    pointRadius: 3,
    pointHoverRadius: 5,
    borderWidth: 3,
    fill: 'origin',
  },
  secondary: {
    backgroundColor: (context: ScriptableContext<'line'>) => getGradient(context, 'secondary'),
    borderColor: '#fc5296',
    pointBackgroundColor: '#fc5296',
    pointHoverBackgroundColor: (context: ScriptableContext<'line'>) =>
      getGradient(context, 'secondary'),
    pointBorderColor: '#fc5296',
    pointHoverBorderColor: (context: ScriptableContext<'line'>) =>
      getGradient(context, 'secondary'),
    pointBorderWidth: 2,
    pointRadius: 3,
    pointHoverRadius: 5,
    borderWidth: 3,
  },
}

interface GraphProps {
  type: string
  variant: string
  title: string
  args: StatisticsParameters
}

export const Graph = ({ type, variant, title, args }: GraphProps) => {
  const [showSettingsModal, setShowSettingsModal] = useState(false)

  const [statistics, setStatistics] = useState<GroupedDatasets>()
  const [previousStatistics, setPreviousStatistics] = useState<GroupedDatasets>()

  const [fromDate, setFromDate] = useState(dayjs().subtract(6, 'day').toDate())
  const [toDate, setToDate] = useState(new Date())

  const [comparePreviousValue, setComparePreviousValue] = useState<ResourceOption | null>(null)
  const [previousFromDate, setPreviousFromDate] = useState<Date>(fromDate)
  const [previousToDate, setPreviousToDate] = useState<Date>(toDate)

  const data = useStatistics({
    ...args,
    timeType: getFilterTimeType(fromDate, toDate),
    fromDate: dayjs(fromDate).format('YYYY-MM-DD'),
    toDate: dayjs(toDate).format('YYYY-MM-DD'),
  })

  const previousData = useStatistics({
    ...args,
    timeType: getFilterTimeType(fromDate, toDate),
    fromDate: dayjs(previousFromDate).format('YYYY-MM-DD'),
    toDate: dayjs(previousToDate).format('YYYY-MM-DD'),
  })

  useEffect(() => {
    if (data.data) {
      const groupedStatistics = groupStatisticsByDate(data.data, fromDate, toDate)
      setStatistics(groupedStatistics)

      if (comparePreviousValue && previousData.data) {
        setShowSettingsModal(false)

        const previousGroupedStatistics = groupStatisticsByDate(
          previousData.data,
          previousFromDate,
          previousToDate,
        )
        setPreviousStatistics(previousGroupedStatistics)
      }
    }
  }, [data.data, previousData.data, comparePreviousValue])

  const datasets = {
    labels: statistics?.labels,
    datasets: [
      {
        data: statistics?.count,
        label: statistics?.labels,
        ...DATASETS_SETTINGS[variant],
        ...(comparePreviousValue && { fill: 'none' }),
      },
      {
        data: comparePreviousValue && previousStatistics ? previousStatistics?.count : [],
        label: previousStatistics?.labels,
        ...DATASETS_SETTINGS['secondary'],
      },
    ],
  }

  //* Set compare previous period dates, prefetches last month
  useEffect(() => {
    setPreviousFromDate(
      dayjs(fromDate)
        .subtract(1, (comparePreviousValue?.value ?? 'month') as ManipulateType)
        .toDate(),
    )
    setPreviousToDate(
      dayjs(toDate)
        .subtract(1, (comparePreviousValue?.value ?? 'month') as ManipulateType)
        .toDate(),
    )
  }, [comparePreviousValue])

  //* Remove compare previous period every time user changes date
  useEffect(() => {
    setComparePreviousValue(null)
  }, [fromDate, toDate])

  return (
    <div className="col-sm-12 col-md-12 col-lg-6 col-xl-6">
      <div className={`card overflow-hidden ${data.isLoading && 'loading-skeleton'}`}>
        {!data.isLoading && (data.isError || !data.data) && (
          <CardError message={data.error?.message} />
        )}
        <div className="card-header d-flex flex-wrap">
          <div className="card-title mb-3 mb-sm-0">
            <span>{title}</span>{' '}
            {comparePreviousValue && (
              <span className="ms-3 badge bg-pink fs-12 fw-bold align-middle">
                {dayjs(previousFromDate).format('DD/MM/YYYY')}
                {' - '}
                {dayjs(previousToDate).format('DD/MM/YYYY')}
              </span>
            )}
          </div>
          <div className="ms-auto">
            <DatePicker
              fromDate={fromDate}
              toDate={toDate}
              onChange={dates => {
                if (Array.isArray(dates) && dates.length === 2) {
                  setFromDate(dates[0])
                  setToDate(dates[1])
                }
              }}
            />
            <button
              className="ms-2 btn btn-icon btn-primary btn-sm"
              onClick={() => setShowSettingsModal(true)}
              aria-label={`Show settings modal for ${title}`}
            >
              <Settings className="w-80 align-middle" />
            </button>
            <GraphSettingsModal
              show={showSettingsModal}
              setShow={(show: boolean) => setShowSettingsModal(show)}
              title={title}
              timeType={getFilterTimeType(fromDate, toDate)}
              valueComparePeriod={comparePreviousValue}
              onComparePeriodChange={option => setComparePreviousValue(option)}
            />
          </div>
        </div>
        <div className="card-body skeleton-img-lg">
          <Chart
            height={330}
            type={type as ChartType}
            options={CHART_OPTIONS}
            data={datasets as ChartData}
            className="chart-dropshadow"
          />
        </div>
      </div>
    </div>
  )
}
